/*
 * kgraft_patch_bsc980883
 *
 * Fix for CVE-2016-4565, bsc#980883, ib_ucm module
 *
 *  Upstream commit:
 *  e6bd18f57aad ("IB/security: Restrict use of the write() interface")
 *
 *  SLE12 commit:
 *  47463e845fb78e44ff63447c8c575aa4737bf7ed
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/fs.h>
#include <rdma/ib_user_cm.h>
#include <asm/uaccess.h>

#include "local.h"

#define KGR_PATCHED_MODULE	"ib_ucm"

static ssize_t (*(*kgr_ucm_cmd_table)[])(struct ib_ucm_file *file, const char __user *inbuf, int in_len, int out_len);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "ucm_cmd_table", (void *)&kgr_ucm_cmd_table },
};

/* patched function */
ssize_t kgr_ib_ucm_write(struct file *filp, const char __user *buf, size_t len,
	loff_t *pos)
{
	struct ib_ucm_file *file = filp->private_data;
	struct ib_ucm_cmd_hdr hdr;
	ssize_t result;

	/*
	 * Fix CVE-2016-4546
	 *  +next 2 lines
	 */
	if (WARN_ON_ONCE(!kgr_ib_safe_file_access(filp)))
		return -EACCES;

	if (len < sizeof(hdr))
		return -EINVAL;

	if (copy_from_user(&hdr, buf, sizeof(hdr)))
		return -EFAULT;

	/*
	 * kgr: hardcode the array size. It is the same across all maintenance
	 * updates.
	 * -if (hdr.cmd >= ARRAY_SIZE(ucm_cmd_table))
	 */
	if (hdr.cmd >= 18)
		return -EINVAL;

	if (hdr.in + sizeof(hdr) > len)
		return -EINVAL;

	result = (*kgr_ucm_cmd_table)[hdr.cmd](file, buf + sizeof(hdr),
					hdr.in, hdr.out);
	if (!result)
		result = len;

	return result;
}

static int kgr_patch_bsc980883_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc980883_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc980883_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc980883_module_nb = {
	.notifier_call = kgr_patch_bsc980883_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc980883_ibucm_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc980883_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc980883_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc980883_ibucm_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc980883_module_nb);
}
